'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _getTypeParameters = require('./getTypeParameters');

var _getTypeParameters2 = _interopRequireDefault(_getTypeParameters);

var _typeAnnotationIterator = require('./typeAnnotationIterator');

var _typeAnnotationIterator2 = _interopRequireDefault(_typeAnnotationIterator);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var converters = {};

/**
 * Convert a type definition to a typed method call.
 */
function convert(context, path) {
  if (!path) {
    return t.identifier('undefined');
  }
  var loc = path.node.loc;
  var converter = converters[path.type];
  if (!converter) {
    return t.identifier('undefined');
  }
  var result = converter(context, path);
  if (!result) {
    return t.identifier('undefined');
  }
  if (loc) {
    result.loc = loc;
  }
  return result;
}

function qualifiedToMemberExpression(context, path) {
  var current = path;
  var stack = [];
  while (current.type === 'QualifiedTypeIdentifier') {
    stack.unshift(current.node.id);
    current = current.get('qualification');
  }

  var first = current.node;
  var second = stack[0];

  // is this a type or a value?
  var entity = context.getEntity(first.name, path);
  var isType = false;
  var isDirectlyReferenceable = false;
  if (entity) {
    if (entity.isValue) {
      isDirectlyReferenceable = true;
    } else {
      isType = true;
      if (entity.isGlobal) {
        isDirectlyReferenceable = false;
      } else {
        isDirectlyReferenceable = true;
      }
    }
  } else {
    isType = true;
  }

  if (!isDirectlyReferenceable) {

    var args = [t.stringLiteral(first.name), t.stringLiteral(second.name)];
    for (var i = 1; i < stack.length; i++) {
      args.push(t.stringLiteral(stack[i]));
    }

    return context.call.apply(context, ['get'].concat(args));
  } else if (isType) {
    var _args = [first, t.stringLiteral(second.name)];
    for (var _i = 1; _i < stack.length; _i++) {
      _args.push(t.stringLiteral(stack[_i]));
    }

    return context.call.apply(context, ['get'].concat(_args));
  } else {

    var inner = t.memberExpression(first, second);

    for (var _i2 = 1; _i2 < stack.length; _i2++) {
      inner = t.memberExpression(inner, stack[_i2]);
    }
    return inner;
  }
}

function annotationToValue(context, subject) {
  switch (subject.type) {
    case 'NullableTypeAnnotation':
    case 'TypeAnnotation':
      return annotationToValue(context, subject.get('typeAnnotation'));
    case 'GenericTypeAnnotation':
      return annotationToValue(context, subject.get('id'));
    case 'QualifiedTypeIdentifier':
      return qualifiedToMemberExpression(context, subject);
    case 'NullLiteralTypeAnnotation':
      return t.nullLiteral();
    case 'VoidTypeAnnotation':
      return t.identifier('undefined');
    case 'BooleanLiteralTypeAnnotation':
      return t.booleanLiteral(subject.node.value);
    case 'NumericLiteralTypeAnnotation':
      return t.numericLiteral(subject.node.value);
    case 'StringLiteralTypeAnnotation':
      return t.stringLiteral(subject.node.value);

    default:
      return subject.node;
  }
}

function getMemberExpressionObject(subject) {
  if (subject.type === 'MemberExpression') {
    return getMemberExpressionObject(subject.object);
  } else {
    return subject;
  }
}

converters.DeclareVariable = function (context, path) {
  var id = path.get('id');
  if (id.has('typeAnnotation')) {
    return context.assumeGlobalDataProperty(id.node.name, convert(context, id.get('typeAnnotation')));
  } else {
    return context.assumeGlobalDataProperty(id.node.name);
  }
};

converters.DeclareFunction = function (context, path) {
  return context.assumeGlobalDataProperty(id.node.name);
};

converters.InterfaceDeclaration = function (context, path) {
  var name = path.node.id.name;
  var body = convert(context, path.get('body'));

  if (path.has('extends')) {
    body = t.objectExpression([].concat(_toConsumableArray(path.get('extends').map(function (item) {
      return convert(context, item);
    }).filter(Boolean).map(function (item) {
      return t.spreadProperty(item);
    })), [t.spreadProperty(body)]));
  }

  return t.variableDeclaration('const', [t.variableDeclarator(t.identifier(name), body)]);
};

converters.InterfaceExtends = function (context, path) {
  var id = path.get('id');
  if (id.isQualifiedTypeIdentifier()) {
    return qualifiedToMemberExpression(context, id);
  } else {
    return t.identifier(id.node.name);
  }
};

converters.DeclareClass = function (context, path) {
  var id = path.get('id');
  var name = id.node.name;

  var body = convert(context, path.get('body'));

  if (path.has('extends')) {
    var parents = path.get('extends').map(function (item) {
      return t.spreadProperty(convert(context, item));
    });

    body = context.abstract(t.objectExpression.apply(t, _toConsumableArray(parents).concat([t.spreadProperty(body)])));
  } else {
    body = context.abstract(body);
  }

  return context.assumeDataProperty(t.memberExpression(t.identifier('global'), t.identifier(name)), 'prototype', body);
};

converters.TypeAlias = function (context, path) {
  var name = path.node.id.name;
  var typeParameters = (0, _getTypeParameters2.default)(path);
  var body = convert(context, path.get('right'));
  return t.variableDeclaration('const', [t.variableDeclarator(t.identifier(name), body)]);
};

converters.TypeofTypeAnnotation = function (context, path) {
  var value = annotationToValue(context, path.get('argument'));
  if (value.type === 'CallExpression') {
    // this is a reference to a type
    return value;
  } else {
    return context.abstract(t.unaryExpression('typeof', value));
  }
};

converters.TypeParameter = function (context, path) {
  // not supported for now.
  return t.identifier('undefined');
};

converters.TypeAnnotation = function (context, path) {
  return convert(context, path.get('typeAnnotation'));
};

converters.NullableTypeAnnotation = function (context, path) {
  // not supported for now.
  return convert(context, path.get('typeAnnotation'));
};

converters.NumberTypeAnnotation = function (context, _ref) {
  var node = _ref.node;

  return t.stringLiteral('number');
};

converters.NumericLiteralTypeAnnotation = function (context, _ref2) {
  var node = _ref2.node;

  return t.stringLiteral('number');
};

// Duplicated for compatibility with flow-parser.
converters.NumberLiteralTypeAnnotation = function (context, _ref3) {
  var node = _ref3.node;

  return t.stringLiteral('number');
};

converters.BooleanTypeAnnotation = function (context, _ref4) {
  var node = _ref4.node;

  return t.stringLiteral('boolean');
};

converters.BooleanLiteralTypeAnnotation = function (context, _ref5) {
  var node = _ref5.node;

  return t.stringLiteral('boolean');
};

converters.StringTypeAnnotation = function (context, _ref6) {
  var node = _ref6.node;

  return t.stringLiteral('string');
};

converters.StringLiteralTypeAnnotation = function (context, _ref7) {
  var node = _ref7.node;

  return t.stringLiteral('string');
};

converters.GenericTypeAnnotation = function (context, path) {
  var id = path.get('id');
  var name = void 0;
  var subject = void 0;
  if (id.isQualifiedTypeIdentifier()) {
    subject = qualifiedToMemberExpression(context, id);
    var outer = getMemberExpressionObject(subject);
    name = outer.name;
  } else {
    name = id.node.name;
    subject = t.identifier(name);
  }
  if (context.shouldSuppressTypeName(name)) {
    return;
  }
  if (context.inTDZ(id.node)) {
    return;
  }
  var typeParameters = (0, _getTypeParameters2.default)(path).map(function (item) {
    return convert(context, item);
  });
  var entity = context.getEntity(name, path);

  if (!entity) {
    if (name === 'Object') {
      return t.stringLiteral('object');
    }
    return;
  }

  var isDirectlyReferenceable = entity && entity.isTypeAlias;

  if (isDirectlyReferenceable) {
    return subject;
  }
};

converters.ObjectTypeAnnotation = function (context, path) {

  var body = path.get('properties').filter(function (property) {
    return !property.node.static;
  });
  return t.objectExpression(body.map(function (item) {
    return convert(context, item);
  }));
};

converters.ObjectTypeSpreadProperty = function (context, path) {
  var arg = convert(context, path.get('argument'));
  if (arg) {
    return t.spreadProperty(arg);
  }
};

converters.ObjectTypeProperty = function (context, path) {
  var propName = void 0;
  if (path.node.optional) {
    // not supported.
    return;
  }
  var value = context.abstract(convert(context, path.get('value')));
  return t.objectProperty(path.node.key, value);
};

exports.default = convert;